/*----------------------------------------------------------------------------------
*
* COPYRIGHT (c) 2001 by Singing Electrons, Inc. All rights reserved.
*
* Module Name			: C:\se\adi\hidclass\Source\HidClass.c
*
* Description			: The HID Class implementation - contains routines that implment the HID Class device
*
*
* Revision History	: At bottom of the file.
*
*---------------------------------------------------------------------------------*/

/** include files **/
#include <sys/exception.h>
#include <sys/excause.h>
//#include <def21535.h>
#include <defBF535.h>
#include "dtype.h"
#include "usb.h"
#include "usbdriver.h"
#include "audioclass.h"
#include "ac97.h"
#include "dprintf.h"

/** local definitions **/
#define PROVIDE_FEEDBACK 1


/* default settings */
#define CONTROL_PIPE_MAX_TRANSFER_LENGTH 32

#define INTERFACE_FOR_AUDIOCONTROL 0
#define ALTSETTING_FOR_AUDIOCONTROL 0

#define INTERFACE_FOR_AUDIOSTREAMING 1
#define ALTSETTING_FOR_AUDIOSTREAMING_IDLE 0
#define ALTSETTING_FOR_AUDIOSTREAMING_48K 1


#define PLAYBACK_EP 0x01
#define PLAYBACK_FEEDBACK_EP 0x82

#define TERMID_USBOUT 0x01
#define UNITID_OUTVOLUME 0x02
#define TERMID_LINEOUT 0x03

#define PLAYBACK_DELAY 3 //milliseconds

//ToDo: These need to be changed based on the codec used
#define OUTVOLUME_MIN 0xa180 //-94.5dB
#define OUTVOLUME_MAX 0x0000 //0 dB
#define OUTVOLUME_RES 0x0180 //1.5 dB


#define NUM_CHANNELS 2
#define CHANNELCONFIG 0x0003 //left and right
#define SAMPLINGRATE 48000

//To create a 24bit audio device change
//the following two values to 24, and also
//change the code that takes data from the buffer 
#define BIT_DEPTH 16
#define CONTAINER_SIZE 16

#if PROVIDE_FEEDBACK
#define FEEDBACK_RANGE 1000 //samples per second
#else
#define FEEDBACK_RANGE 0 //samples per second
#endif

#define MAX_PACKET_SIZE (((SAMPLINGRATE + FEEDBACK_RANGE)  * NUM_CHANNELS * (CONTAINER_SIZE / 8)) / 1000)
#define FEEDBACK_PACKET_SIZE 3

//#define NUM_PACKETS_TO_BUFFER 7
//#define MAX_DATA_BUFFER_SIZE (MAX_PACKET_SIZE * NUM_PACKETS_TO_BUFFER)

//0000110000.0000000000  = 48.0 (48 + 0/1024) (bcd 10.10 format)
//000011000000000000000000  = 48.0 (48 + 0/1024) = 0c0000
#define FEEDBACK_FREQUENCY_NORMAL 0x0c0000

//0000110000.1000000000 = 48.5 (48 + 512/1024)
//000011000010000000000000  = 48.5 (48 + 512/1024) = 0c2000
//#define FEEDBACK_FREQUENCY_PLEASE_HURRY 0x0c2000
#define FEEDBACK_FREQUENCY_PLEASE_HURRY 0x0c4000 //49K


//0000101111.1000000000  = 47.5 (47 + 512/1024)
//000010111110000000000000  = 47.5 (47 + 512/1024) = 0be000
//#define FEEDBACK_FREQUENCY_PLEASE_SLOWDOWN 0x0be000
#define FEEDBACK_FREQUENCY_PLEASE_SLOWDOWN 0x0bc000 //47K


#include "descriptors.h"



/** external functions **/

/** external data **/

/** internal functions **/

/** public data **/
bool gIsPlaybackActive;


/** private data **/
static PSUSPENDPROC gpOnSuspend;
static PRESUMEPROC gpOnResume;
static PRESETPROC gpOnReset;

static UCHAR gOutMute;
static USHORT gOutLeftVol;
static USHORT gOutRightVol;

static UINT gFeedbackData;
static USHORT gEnablePlaybackCountdown;
static UINT gRegChangesRequested, gRegChangesImplemented;

static UINT gTimeoutCount;

//static UINT gDataBufferHead;
//static UINT gDataBufferTail;
//static UCHAR gDataBuffer[MAX_DATA_BUFFER_SIZE];

static UINT gDebugNumTransfers;
static UINT gDebugTotalTransferLength;

/** public functions **/

/** private functions **/
bool AudioGetRequest (UCHAR bRequest, UCHAR wIndexHi, USHORT wValue, USHORT wLength, PUINT pNumBytesToTransfer, UCHAR *buffer);
bool AudioSetRequest (UCHAR bRequest, UCHAR wIndexHi, USHORT wValue, USHORT wLength, UCHAR *buffer);

/** callback functions **/
static bool OnSetupRequest (UCHAR endpointNumber, PUSB_SETUP_DATA pSetupData, void *buffer, 
	PUINT pNumBytesToTransfer);
static bool OnGetDeviceDescriptor (void *buffer, PUINT pNumBytesToTransfer);
static bool OnGetConfigDescriptor (void *buffer, PUINT pNumBytesToTransfer);
static bool OnGetStringDescriptor (UCHAR stringIndex, USHORT langID, void *buffer, PUINT pNumBytesToTransfer);
static void OnInTransferComplete (UCHAR endpointNumber);
static bool OnOutTransferComplete (UCHAR endpointNumber, void *buffer, UINT numBytesReceived, bool wasPCAsserted);
static void OnNAKSent (UCHAR endpointNumber);
static bool OnConfigRequest (UCHAR newConfiguration, UCHAR newInterface, UCHAR newAltSetting);
static void OnSuspend (void);
static void OnResume (void);
static void OnReset (void);
static void OnSOF (void);




/*------------------------------------------------------------
*	HIDCLASS_Init
*
*	Parameters:  
*
*	Globals Used:
*		gpOnResume, gpOnReset, gUseINEP, gpTheReportDescriptor, gTheReportDescriptorSize
*		gTheDeviceDescriptor, gTheConfigDescriptor
*
*	Description:
*		This is the main initialization routine. The client provides the vendor id, device id,
*		revision, and a bunch of pointers to routines to be calleld under various conditions.
*
*	Returns:
*		true on success, false on failure.
*
*------------------------------------------------------------*/
bool AUDIOCLASS_Init (uint16 vid, uint16 pid, uint16 rel, PSUSPENDPROC pOnSuspendProc, 
	PRESUMEPROC pOnResumeProc, PRESETPROC pOnResetProc)
	{
		if (!AC97_Init())
			return false;
			
		//Initialize the state variables
		gOutMute = 0;
		gOutLeftVol = OUTVOLUME_MAX;
		gOutRightVol = OUTVOLUME_MAX;
		gIsPlaybackActive = false;
		gEnablePlaybackCountdown = 0;
		//gDataBufferHead = gDataBufferTail = 0;
		//memset (&gDataBuffer, 0, sizeof (gDataBuffer));

		gDebugNumTransfers = 0;
		gDebugTotalTransferLength = 0;

		gFeedbackData = FEEDBACK_FREQUENCY_NORMAL;

		//Save the callback addresses
		gpOnSuspend = pOnSuspendProc;
		gpOnResume = pOnResumeProc;
		gpOnReset = pOnResetProc;

		gRegChangesRequested = gRegChangesImplemented = 0;


		//Update the VID and PID
		memcpy (&gTheDeviceDescriptor[OFFSET_VENDOR_ID], &vid, sizeof (uint16));
		memcpy (&gTheDeviceDescriptor[OFFSET_PRODUCT_ID], &pid, sizeof (uint16));
		memcpy (&gTheDeviceDescriptor[OFFSET_RELEASE], &rel, sizeof (uint16));


		//Things to do before setting up endpoints
		if (!USBDriver_PreInit ())
			return false;

		USBDriver_SetupEndpoint (EP1, ALTSETTING_FOR_AUDIOSTREAMING_IDLE, 
			INTERFACE_FOR_AUDIOSTREAMING, EP_CONFIG1, EP_OUT,
			EP_ISO, 0);

		USBDriver_SetupEndpoint (EP1, ALTSETTING_FOR_AUDIOSTREAMING_48K, 
			INTERFACE_FOR_AUDIOSTREAMING, EP_CONFIG1, EP_OUT,
			EP_ISO, MAX_PACKET_SIZE);

#if PROVIDE_FEEDBACK		
		USBDriver_SetupEndpoint (EP2, ALTSETTING_FOR_AUDIOSTREAMING_48K, 
			INTERFACE_FOR_AUDIOSTREAMING, EP_CONFIG1, EP_IN,
			EP_ISO, FEEDBACK_PACKET_SIZE);
#endif //PROVIDE_FEEDBACK		

		return USBDriver_CompleteInit (OnSetupRequest, OnGetDeviceDescriptor, 
			OnGetConfigDescriptor, OnGetStringDescriptor, 
			OnInTransferComplete, OnOutTransferComplete, OnNAKSent, 
			OnSuspend, OnResume, OnReset, OnConfigRequest, OnSOF, CONTROL_PIPE_MAX_TRANSFER_LENGTH);

	}



/*------------------------------------------------------------
*	AUDIOCLASS_DoForegroundTasks
*
*	Parameters:  None
*
*	Globals Used:
*		gRegChangesImplemented, gRegChangesRequested
*
*	Description:
*		This routine essentially does the time consuming task
*		of talking to the AC97 codec when requested for volume
*		and mute controls. This way the ISR does not have to
*		waste a lot of time.
*
*	Returns:
*		Nothing
*
*------------------------------------------------------------*/
void AUDIOCLASS_DoForegroundTasks (void)
	{
		if (gRegChangesRequested != gRegChangesImplemented)
			{
				AC97_SetVolumeAndMute (gOutLeftVol, gOutRightVol, gOutMute);
				gRegChangesImplemented++;
			}
	}



/*------------------------------------------------------------
*	OnSetupRequest
*
*	Parameters:  
*		endpointNumber - the endpoint on which this request was received
*		pSetupData - points to a structure cantaining the SETUP DATA(0)
*		buffer - points to a buffer where the setup data is available, or should be placed.
*		pNumBytesToTransfer - points to a location where the number of bytes
*			to be sent back should be stored.
*	Globals Used:
*		None
*
*	Description:
*		This routine is called by the USB driver to handle a special setup request, 
*		that is not a GET DEVICE, GET CONFIG, or GET STRING descriptor
*		
*		Based on the request type, this routine reads/writes to the buffer
*		provided.
*
*	Returns:
*		true to succeed the request, false to stall the request
*
*------------------------------------------------------------*/
bool OnSetupRequest (UCHAR endpointNumber, PUSB_SETUP_DATA pSetupData, void *buffer, 
	PUINT pNumBytesToTransfer)
	{
		bool retVal = false; //stall
		bool validRequest = false;

		//Need to handle the requests used to change the various controls
		if (endpointNumber == EP0)
			{
				if ((USB_CONTROL_REQUEST_GET_TYPE (pSetupData->bmRequestType) == USB_CONTROL_REQUEST_TYPE_CLASS) &&
					(USB_CONTROL_REQUEST_GET_RECIPIENT (pSetupData->bmRequestType) == USB_CONTROL_REQUEST_TYPE_INTERFACE) &&
					(LOBYTE (pSetupData->wIndex) == INTERFACE_FOR_AUDIOCONTROL))
					{
						validRequest = true;
						if (USB_CONTROL_REQUEST_GET_DIRECTION(pSetupData->bmRequestType) == USB_CONTROL_REQUEST_TYPE_OUT)
							retVal = AudioSetRequest (pSetupData->bRequest, HIBYTE(pSetupData->wIndex), pSetupData->wValue, pSetupData->wLength, buffer);
						else
							retVal = AudioGetRequest (pSetupData->bRequest, HIBYTE(pSetupData->wIndex), pSetupData->wValue, pSetupData->wLength, pNumBytesToTransfer, buffer);
					}
			}

		
		if (!validRequest)
			{
				Dprintf1 ("Unknown setup request: %02x", pSetupData->bmRequestType);
				Dprintf3 ("Dir = %02x, Type = %02x, Rec = %02x", USB_CONTROL_REQUEST_GET_DIRECTION (pSetupData->bmRequestType),
					USB_CONTROL_REQUEST_GET_TYPE (pSetupData->bmRequestType),
					USB_CONTROL_REQUEST_GET_RECIPIENT (pSetupData->bmRequestType));
				Dprintf1 ("bRequest = %02x", pSetupData->bRequest);
			}
			

		return (retVal);
	}


/*------------------------------------------------------------
*	AudioGetRequest
*
*	Parameters:  
*		bRequest - the request.
*		wIndexHi - the high byte of wIndex
*		wValue - as it comes in the setup packet
*		wLength - as it comes in the setup packet.
*		pNumBytesToTransfer - pointer to a location which receives the number of bytes to be transferred.
*		buffer - pointer to a buffer where the result data is to be stored.
*
*	Globals Used:
*		gOutMute, gOutLeftVol, gOutRightVol
*
*	Description:
*		This handles the Audio Class Get Requests
*
*	Returns:
*		true on success, false on failure (to stall the EP).
*
*------------------------------------------------------------*/
bool AudioGetRequest (UCHAR bRequest, UCHAR wIndexHi, USHORT wValue, USHORT wLength, PUINT pNumBytesToTransfer, UCHAR *buffer)
	{
		bool retVal = false;

		switch (bRequest)
			{
				case GET_CUR:
					switch (wIndexHi)
						{
							case UNITID_OUTVOLUME:
								switch (HIBYTE(wValue))
									{
										case USB_AUDIO_CONTROL_SELECTOR_MUTE_CONTROL:
											if (LOBYTE(wValue) == 0)
												{
													buffer[0] = gOutMute;
													*pNumBytesToTransfer = sizeof (UCHAR);
													retVal = true;
												}
											break;
										case USB_AUDIO_CONTROL_SELECTOR_VOLUME_CONTROL:
											switch (LOBYTE(wValue))
												{
													case 0:
														buffer[0] = LOBYTE ((gOutLeftVol + gOutRightVol) >> 1);
														buffer[1] = HIBYTE ((gOutLeftVol + gOutRightVol) >> 1);
														*pNumBytesToTransfer = sizeof (USHORT);
														retVal = true;
														break;
													case 1:
														buffer[0] = LOBYTE (gOutLeftVol);
														buffer[1] = HIBYTE (gOutLeftVol);
														*pNumBytesToTransfer = sizeof (USHORT);
														retVal = true;
														break;
													case 2:
														buffer[0] = LOBYTE (gOutRightVol);
														buffer[1] = HIBYTE (gOutRightVol);
														*pNumBytesToTransfer = sizeof (USHORT);
														retVal = true;
														break;
													case 0xff:
														buffer[0] = LOBYTE ((gOutLeftVol + gOutRightVol) >> 1);
														buffer[1] = HIBYTE ((gOutLeftVol + gOutRightVol) >> 1);
														buffer[2] = LOBYTE (gOutLeftVol);
														buffer[3] = HIBYTE (gOutLeftVol);
														buffer[4] = LOBYTE (gOutRightVol);
														buffer[5] = HIBYTE (gOutRightVol);
														*pNumBytesToTransfer = sizeof (USHORT) * 3;
														retVal = true;
														break;
													default:
														//Shouldn't really get here!
														break;
												}
											break;
									}
								break;
							default:
								//Shouldn't really get here?
								break;
						}
					break;
				case GET_MIN:
					switch (wIndexHi)
						{
							case UNITID_OUTVOLUME:
								buffer[0] = LOBYTE(OUTVOLUME_MIN);
								buffer[1] = HIBYTE(OUTVOLUME_MIN);
								if (LOBYTE(wValue) == 0xff)
									{
										memcpy (buffer+2, buffer, 2);
										memcpy (buffer+4, buffer, 2);
										*pNumBytesToTransfer = 6;
									}
								else
									*pNumBytesToTransfer = 2;
								retVal = true;
								break;
							default:
								//Shouldn't really get here?
								break;
						}
					break;
				case GET_MAX:
					switch (wIndexHi)
						{
							case UNITID_OUTVOLUME:
								buffer[0] = LOBYTE(OUTVOLUME_MAX);
								buffer[1] = HIBYTE(OUTVOLUME_MAX);
								if (LOBYTE(wValue) == 0xff)
									{
										memcpy (buffer+2, buffer, 2);
										memcpy (buffer+4, buffer, 2);
										*pNumBytesToTransfer = 6;
									}
								else
									*pNumBytesToTransfer = 2;
								retVal = true;
								break;
							default:
								//Shouldn't really get here?
								break;
						}
					break;
				case GET_RES:
					switch (wIndexHi)
						{
							case UNITID_OUTVOLUME:
								buffer[0] = LOBYTE(OUTVOLUME_RES);
								buffer[1] = HIBYTE(OUTVOLUME_RES);
								if (LOBYTE(wValue) == 0xff)
									{
										memcpy (buffer+2, buffer, 2);
										memcpy (buffer+4, buffer, 2);
										*pNumBytesToTransfer = 6;
									}
								else
									*pNumBytesToTransfer = 2;
								retVal = true;
								break;
							default:
								//Shouldn't really get here?
								break;
						}
					break;
				default:
					//Shouldn't really get here!
					break;
			}

		return(retVal);
		
	}


/*------------------------------------------------------------
*	AudioSetRequest
*
*	Parameters:  
*		bRequest - the request.
*		wIndexHi - the high byte of wIndex
*		wValue - as it comes in the setup packet
*		wLength - as it comes in the setup packet.
*		buffer - pointer to a buffer where the associated data is stored
*
*	Globals Used:
*		gOutMute, gOutLeftVol, gOutRightVol
*
*	Description:
*		This handles the Audio Class Set Requests
*
*	Returns:
*		true on success, false on failure (to stall the EP).
*
*------------------------------------------------------------*/
bool AudioSetRequest (UCHAR bRequest, UCHAR wIndexHi, USHORT wValue, USHORT wLength, UCHAR *buffer)
	{
		bool retVal = false;

		//Dprintf3 ("wIndexHi = %02x, wValue = %04x, wData = %04x", wIndexHi, wValue, *(PUSHORT)buffer);

		switch (bRequest)
			{
				case SET_CUR:
					switch (wIndexHi)
						{
							case UNITID_OUTVOLUME:
								switch (HIBYTE(wValue))
									{
										case USB_AUDIO_CONTROL_SELECTOR_MUTE_CONTROL:
											if (LOBYTE(wValue) == 0)
												{
													gOutMute = buffer[0];
													//Call whatever is required to mute the codec
													gRegChangesRequested++;
													retVal = true;
													break;
												}
											break;
										case USB_AUDIO_CONTROL_SELECTOR_VOLUME_CONTROL:
											switch (LOBYTE(wValue))
												{
													case 0:
														memcpy (&gOutLeftVol, buffer, sizeof (USHORT));
														memcpy (&gOutRightVol, buffer, sizeof (USHORT));
														//Call whatever is required to change the codec volume
														gRegChangesRequested++;
														retVal = true;
														break;
													case 1:
														memcpy (&gOutLeftVol, buffer, sizeof (USHORT));
														//Call whatever is required to change the codec volume
														gRegChangesRequested++;
														retVal = true;
														break;
													case 2:
														memcpy (&gOutRightVol, buffer, sizeof (USHORT));
														//Call whatever is required to change the codec volume
														gRegChangesRequested++;
														retVal = true;
														break;
													case 0xff:
														memcpy (&gOutLeftVol, buffer + sizeof(USHORT), sizeof (USHORT));
														memcpy (&gOutRightVol, buffer + 2 * (sizeof(USHORT)), sizeof (USHORT));
														//Call whatever is required to change the codec volume
														gRegChangesRequested++;
														retVal = true;
														break;
													default:
														//Shouldn't really get here!
														break;
												}
											break;
									}
								
								break;
							default:
								//Shouldn't really get here?
								break;
						}
					break;
				default:
					//Should not really get here
					break;
			}
		return(retVal);
	}


/*------------------------------------------------------------
*	OnGetDeviceDescriptor
*
*	Parameters:  
*		buffer - location (DMA buffer) where the descriptor data is to
*			be put.
*		pNumBytesToTransfer - points to a location where the original wLength
*			is stored. This routine should update this location to indicate
*			the actual number of bytes transferred to the buffer.
*
*	Globals Used:
*		gTheDeviceDescriptor
*
*	Description:
*		This routine is in response to a GET DEVICE DESCRIPTOR
*		setup request. This routine fills in the required data
*
*	Returns:
*		true - always!
*
*------------------------------------------------------------*/
bool OnGetDeviceDescriptor (void *buffer, PUINT pNumBytesToTransfer)
	{
		if (*pNumBytesToTransfer < sizeof (gTheDeviceDescriptor))
			memcpy (buffer, &gTheDeviceDescriptor, *pNumBytesToTransfer);
		else
			{
				memcpy (buffer, &gTheDeviceDescriptor, sizeof (gTheDeviceDescriptor));
				*pNumBytesToTransfer = sizeof (gTheDeviceDescriptor);
			}

		return(true);
	}



/*------------------------------------------------------------
*	OnGetConfigDescriptor
*
*	Parameters:  
*		buffer - location (DMA buffer) where the descriptor data is to
*			be put.
*		pNumBytesToTransfer - points to a location where the original wLength
*			is stored. This routine should update this location to indicate
*			the actual number of bytes transferred to the buffer.
*
*	Globals Used:
*		gTheConfigDescriptor
*
*	Description:
*		This routine is in response to a GET CONFIG DESCRIPTOR
*		setup request. This routine fills in the required data
*
*	Returns:
*		true - always!
*
*------------------------------------------------------------*/
bool OnGetConfigDescriptor (void *buffer, PUINT pNumBytesToTransfer)
	{

		if (*pNumBytesToTransfer < sizeof(gTheConfigDescriptor))
			memcpy (buffer, &gTheConfigDescriptor, *pNumBytesToTransfer);
		else
			{
				memcpy (buffer, &gTheConfigDescriptor, sizeof (gTheConfigDescriptor));
				*pNumBytesToTransfer = sizeof (gTheConfigDescriptor);
			}

		return (true);
	}



/*------------------------------------------------------------
*	OnGetStringDescriptor
*
*	Parameters:  
*		wValue - the string to be retrieved
*		buffer - location (DMA buffer) where the descriptor data is to
*			be put.
*		pNumBytesToTransfer - points to a location where the original wLength
*			is stored. This routine should update this location to indicate
*			the actual number of bytes transferred to the buffer.
*
*	Globals Used:
*		Not Yet!
*
*	Description:
*		This routine is in response to a GET STRING DESCRIPTOR
*		setup request. This routine fills in the required data
*
*	Returns:
*		true - always!
*
*------------------------------------------------------------*/
bool OnGetStringDescriptor (UCHAR stringIndex, USHORT langID, void *buffer, PUINT pNumBytesToTransfer)
	{
		bool retVal = false;
		UINT neededSpace = 0;
		UINT maxBytesToTransfer = 0;
		UINT outIndex = 0, charIndex = 0;

		if (stringIndex == 0)
			{
				//Need to return a langid array here!
				if (*pNumBytesToTransfer < sizeof (gTheLangIDArray))
					memcpy (buffer, &gTheLangIDArray, *pNumBytesToTransfer);
				else
					{
						memcpy (buffer, &gTheLangIDArray, sizeof (gTheLangIDArray));
						*pNumBytesToTransfer = sizeof (gTheLangIDArray);
					}
				retVal = true;
			}
		else if ((stringIndex < TOTAL_STRING_DESCRIPTORS) && (stringIndex > 0))
			{
				//Convert this to a zero based index
				stringIndex--;
				neededSpace = (strlen (gTheStringDescriptors[stringIndex]) + 1) * 2;
				if (neededSpace > *pNumBytesToTransfer)
					maxBytesToTransfer = *pNumBytesToTransfer;
				else
					maxBytesToTransfer = *pNumBytesToTransfer = neededSpace;

				//Transfer as many as we can
				((char *)buffer)[outIndex++] = neededSpace;
				((char *)buffer)[outIndex++] = USB_STRING_DESCRIPTOR;
				while ((outIndex + 1) < maxBytesToTransfer)
					{
						((char *)buffer)[outIndex++] = gTheStringDescriptors[stringIndex][charIndex++];
						((char *)buffer)[outIndex++] = 0;
					}


				retVal = true;
			}

		//Don't have any strings yet!
		return(retVal);
	}




/*------------------------------------------------------------
*	OnConfigRequest
*
*	Parameters:  
*		newConfiguration - the new configuration after Config request.
*		newInterface - the new interface after Config(Set Interface) request.
*		newAltSetting - the new alt setting after Config(Set Interface) request
*
*	Globals Used:
*		gPrepareReport
*
*	Description:
*		This routien is called in response to a configuration change - i.e.
*		after receiving a Set Config or a Set Interface request.
*
*	Returns:
*		true - always
*
*------------------------------------------------------------*/
bool OnConfigRequest (UCHAR newConfiguration, UCHAR newInterface, UCHAR newAltSetting)
	{
		//Dprintf3 ("Audio Dev Config Request: %02x, %02x, %02x!", (UINT)newConfiguration, (UINT)newInterface, (UINT)newAltSetting);
		if ((newConfiguration == EP_CONFIG1))// && (newInterface == INTERFACE_FOR_HID) &&
			//(newAltSetting == ALTSETTING_FOR_HID))
			{
				//Looks like we're in, setup our IN endpoint if we want to!
				Dprintf ("Audio Device Configured!");
				if (newInterface == INTERFACE_FOR_AUDIOSTREAMING)
					{
						if ((newAltSetting == ALTSETTING_FOR_AUDIOSTREAMING_48K) && (!gIsPlaybackActive))
							{
								gTimeoutCount = 0;
								gIsPlaybackActive = true;
								//ToDo: Initialize the codec to start receiving data.
								Dprintf ("Starting Playback!");


#if PROVIDE_FEEDBACK
								//Initialize Feedback Fs.
								gFeedbackData = FEEDBACK_FREQUENCY_NORMAL;
								//Activate Endpoint
								ActivateEndpoint (EP2, USBD_TYP_ISO, FEEDBACK_PACKET_SIZE, FEEDBACK_PACKET_SIZE);
								//Prepare for a transfer.
								OnInTransferComplete (EP2);

#endif // PROVIDE_FEEDBACK

								//init the pointers
								gDataBufferHead = gDataBufferTail = 0;

								//Setup and Arm our endpoint
								ActivateEndpoint (EP1, USBD_TYP_ISO, MAX_PACKET_SIZE, MAX_PACKET_SIZE);
								PrepareOUTEPForXFer (EP1);
								ArmEndpoint (EP1, EP_OUT);
								gEnablePlaybackCountdown = PLAYBACK_DELAY;
							}
						else if ((newAltSetting == ALTSETTING_FOR_AUDIOSTREAMING_IDLE) && (gIsPlaybackActive))
							{
								Dprintf ("Stopping Playback!");
								gIsPlaybackActive = false;
								//Disarm the endpoint
								DisarmEndpoint (EP1, EP_OUT);
								UnprepareEP (EP1);

#if PROVIDE_FEEDBACK
								DisarmEndpoint (EP2, EP_IN);
								UnprepareEP (EP2);
#endif //PROVIDE_FEEDBACK

								//ToDo: Cleanup the codec so that it stops playing
								AC97_DisablePlayback();
							}
					}
			}
		else
			{
				if (gIsPlaybackActive)
					{
						gIsPlaybackActive = false;
						//Disarm the endpoint
						DisarmEndpoint (EP1, EP_OUT);
						UnprepareEP (EP1);
#if PROVIDE_FEEDBACK
						DisarmEndpoint (EP2, EP_IN);
						UnprepareEP (EP2);
#endif //PROVIDE_FEEDBACK
						//ToDo: Cleanup the codec so that it stops playing
						AC97_DisablePlayback();
					}
				//We should disarm the endpoint(s) at this point!
			}
			
		return true;
	}



/*------------------------------------------------------------
*	OnInTransferComplete
*
*	Parameters:  
*		endpointNumber - the endpoint on which the transfer was completed.
*
*	Globals Used:
*		none
*
*	Description:
*		This routine is called when a IN transfer is completed on one of the
*		endpoints. 
*
*	Returns:
*		Nothing
*
*------------------------------------------------------------*/
void OnInTransferComplete (UCHAR endpointNumber)
	{
		void *pFeedbackData;

#if PROVIDE_FEEDBACK
		if ((endpointNumber == EP2) && (gIsPlaybackActive))
			{
				if (PrepareINEPForXFer (EP2, &pFeedbackData))
					{
						memcpy (pFeedbackData, &gFeedbackData, FEEDBACK_PACKET_SIZE);
						//Arm the endpoint!
						ArmEndpoint (EP2, EP_IN);
					}
			}
#endif //PROVIDE_FEEDBACK
	}


/*------------------------------------------------------------
*	OnOutTransferComplete
*
*	Parameters:  
*		endpointNumber - the endpoint on which the transfer was completed.
*		buffer - data buffer where the data was received
*		numBytesReceived - number of bytes received
*		wasPCAsserted - was the PC(packet complete) asserted when the TC interrup
*			came in. If not, this could be a possible problem???
*
*	Globals Used:
*		gIsPlaybackActive
*
*	Description:
*		This routine is called when a transfer is completed on one of the
*		out endpoints. In turn, this routine saves the received data in
*		a cyclic buffer.
*
*	Returns:
*		true to tell the isr to recycle the dma buffer, and setup
*		things for the next transfer, false to tell the isr to
*		free up the dma buffer, and disarm the endpoint.
*
*------------------------------------------------------------*/
bool OnOutTransferComplete (UCHAR endpointNumber, void *buffer, UINT numBytesReceived, bool wasPCAsserted)
	{

		UINT inThisPass, currentTail, headTailDistance;

		if (!gIsPlaybackActive)
			return (false);

		gTimeoutCount = 0;

		//gDebugNumTransfers++;
		//gDebugTotalTransferLength += numBytesReceived;
			
		if ((gDataBufferHead + numBytesReceived) >= MAX_DATA_BUFFER_SIZE)
			{
				inThisPass = MAX_DATA_BUFFER_SIZE - gDataBufferHead;
				memcpy (&gDataBuffer[gDataBufferHead], buffer, inThisPass);
				memcpy (&gDataBuffer, (UCHAR *)buffer + inThisPass, numBytesReceived - inThisPass);
				gDataBufferHead = numBytesReceived - inThisPass;
			}
		else
			{
				memcpy (&gDataBuffer[gDataBufferHead], buffer, numBytesReceived);
				gDataBufferHead += numBytesReceived;
			}


		if (gEnablePlaybackCountdown != 0)
			{
				gEnablePlaybackCountdown--;
				if (gEnablePlaybackCountdown == 0)
					AC97_EnablePlayback ();
			}
#if PROVIDE_FEEDBACK
		else
			{
				//Check to see if we need to change the feedback Fs?

				//Need to latch this value, as the SPORT0 interrupt
				//might change it any time!
				currentTail = gDataBufferTail;

				if (gDataBufferHead >= currentTail)
					headTailDistance = gDataBufferHead - currentTail;
				else
					headTailDistance = (MAX_DATA_BUFFER_SIZE - currentTail) + gDataBufferHead;
				
				if (headTailDistance > ((PLAYBACK_DELAY + 1) * 4 * 48))
					{
						gFeedbackData = FEEDBACK_FREQUENCY_PLEASE_SLOWDOWN;
						//Dprintf2 ("Slow: Head:%d, Tail: %d", gDataBufferHead, currentTail);
					}
				else if (headTailDistance < ((PLAYBACK_DELAY -1) * 4 * 48))
					{
						gFeedbackData = FEEDBACK_FREQUENCY_PLEASE_HURRY;
						//Dprintf2 ("Hurry: Head:%d, Tail: %d", gDataBufferHead, currentTail);
					}
				else
					gFeedbackData = FEEDBACK_FREQUENCY_NORMAL;

			}
#endif //PROVIDE_FEEDBACK

		return (true);
		

	}


/*------------------------------------------------------------
*	OnNAKSent
*
*	Parameters:  
*		endpointNumber - endpoint where the NAK was sent.
*
*	Globals Used:
*		gPrepareReport
*
*	Description:
*		This routine is called when the device has sent a NAK in response
*		to an IN or OUT request while the DMA buffer for that endpoint
*		has been emptied. Please note that this will only be called in
*		response to an "expected" direction request - IN request for
*		endpoints setup as IN endpoints and OUT request for endpoints
*		setup as OUT endpoints. 
*
*	Returns:
*		Nothing
*
*------------------------------------------------------------*/
void OnNAKSent (UCHAR endpointNumber)
	{
	}




/*------------------------------------------------------------
*	OnSuspend
*
*	Parameters:  
*
*	Globals Used:
*		gpOnSuspend
*
*	Description:
*		This routine will be called when USBDriver detects a
*		suspend condition. Currently, it quites the AC97 and forwards
*		the call to the client!
*
*	Returns:
*		Nothing.
*
*------------------------------------------------------------*/
void OnSuspend (void)
	{
		gIsPlaybackActive = false;
		AC97_DisablePlayback ();
		if (gpOnSuspend)
			(gpOnSuspend)();
	}



/*------------------------------------------------------------
*	OnResume
*
*	Parameters:  
*
*	Globals Used:
*		gpOnResume
*
*	Description:
*		This routine will be called when the USBDriver detects that
*		the bus is out of suspend. Currently, it simply forwards
*		the call to the client!
*
*	Returns:
*		Nothing.
*
*------------------------------------------------------------*/
void OnResume (void)
	{
		if (gpOnResume)
			(gpOnResume) ();
	}


/*------------------------------------------------------------
*	OnSOF
*
*	Parameters:  None.
*
*	Globals Used:
*		gIsPlaybackActive, gTimeoutCount, gDataBufferHead, gDataBufferTail
*		gEnablePlaybackCountdown
*
*	Description:
*		This routine will be called when the USBDriver receives an SOF
*		interrupt. It is mainly used to see if the transfers have timed out.
*
*	Returns:
*		Nothing.
*
*------------------------------------------------------------*/
void OnSOF (void)
	{
		if (!gIsPlaybackActive)
			return;

		if (gTimeoutCount > 3)
			{
				if (gEnablePlaybackCountdown == 0)
					{
						AC97_DisablePlayback ();
						//init the pointers
						gDataBufferHead = gDataBufferTail = 0;
						gEnablePlaybackCountdown = PLAYBACK_DELAY;
					}
			}
		else
			gTimeoutCount++;

	}


/*------------------------------------------------------------
*	OnReset
*
*	Parameters:  
*
*	Globals Used:
*		gpOnReset
*
*	Description:
*		This routine will be called when the USBDriver detects reset
*		signalling on the bus. Currently, it quiets the codec and forwards
*		the call to the client!
*
*	Returns:
*		Nothing.
*
*------------------------------------------------------------*/
void OnReset (void)
	{
		gIsPlaybackActive = false;
		AC97_DisablePlayback ();
		if (gpOnReset)
			(gpOnReset)();
	}


/*----------------------------------------------------------------------------------
* $Log: AudioClass.c,v $
* Revision 1.8  2003/01/17 00:53:10  Devendra
* Commented out a few Dprintfs
*
* Revision 1.7  2003/01/17 00:52:03  Devendra
* - BugFix: Set Volume requests were not handled properly.
* - Changed the feedback speed up and slow down rates to be more aggressive.
*
* Revision 1.6  2003/01/16 19:10:05  Devendra
* - Added code to set AC97 registers for mute and volume controls based on Host requests..
* - Added code to detect "timeout" (no isoch. packets coming in) by implementing OnSOF callback.
* - Added support for AudioClass to perform non-ISR(foreground) tasks.
* - Exposed playback active state variable.
*
* Revision 1.5  2003/01/15 01:35:14  Devendra
* Added code to provide feedback to pace control  the isoch. pipe.
*
* Revision 1.4  2003/01/13 19:56:18  Devendra
* Moved the data buffer to ac97.c
* Added code to enuser "distance" between buffer head and tail (latency).
* BugFix: buffer copy was not proper.
* Separated Descriptors to a separate header.
*
* Revision 1.2  2003/01/10 01:53:55  Devendra
* - Added support for receiving a playback stream on isoch. endpoint and storing the data in a cyclic buffer.
*
* Revision 1.1  2003/01/09 01:16:35  Devendra
* First Rev of AudioClass - have the device enumerated as audio device, and the volume control and mute are functional!
*
* Revision 1.1  2002/10/31 00:30:59  Devendra
* Moved all files in one folder to avoid IDE related problems.
*
* Revision 1.3  2002/10/30 02:36:40  Devendra
* - Added more abstraction to HIDCLASS. So that it's possible to build
*   different devices without making any changes to the HIDCLASS module.
* - Added proper support for handling USB SUSPEND, RESUME, and RESET events.
*
* Revision 1.2  2002/10/22 17:50:54  Devendra
* - Changed Product ID to 0x0005
*
* Revision 1.1  2002/10/22 17:23:38  Devendra
* Rearranged file locations.
*
* Revision 1.6  2002/10/14 05:16:48  Devendra
* Updated version number.
*
* Revision 1.5  2002/10/14 05:16:04  Devendra
* - Custom Device now using Feature reports for I/O
* - Added compiler define/project configuration to switch between mouse and custom device.
*
* Revision 1.4  2002/10/14 01:33:15  Devendra
* Got "custom" device functional with buttons and lights!
*
* Revision 1.3  2002/10/09 17:09:32  Devendra
* - Added support for handling IN endpoints.
* - Added HID descriptors (HID and HID Report)
* - Modified the Interface descriptors to comply to HID Class Spec.
* - Added functionality to use the push-buttons on the Eagle-35 as a mouse (for buttons as well as for movement).
* - The device is now a fully functioning USB Mouse!
*
* Revision 1.2  2002/10/08 19:57:02  Devendra
* - Added specific callbacks for getting Device, Config, and String descriptors.
* - BugFix: Max. Packet Size was not being set correctly in ArmEndpoint.
* - Added internal buffer based logging to allow real-time "printfs".
* - Added String descriptors.
*
* Revision 1.1  2002/09/20 06:51:30  Devendra
* First Rev.
*
*
*---------------------------------------------------------------------------------*/
