/*----------------------------------------------------------------------------------
*
* COPYRIGHT (c) 2001 by Singing Electrons, Inc. All rights reserved.
*
* This document contains confidential proprietary trade secrets of Singing Electrons, Inc.  
* No distribution of the contents of this file is allowed except by written permission by an officer 
* of Singing Electrons, Inc.
*
* Module Name			: C:\se\adi\hidclass\source\hosttest.c
*
* Description			: A test application to test the generic HID class device.
*
*
* Revision History	: At bottom of the file.
*
*---------------------------------------------------------------------------------*/

/** include files **/
#include <windows.h>
#include <setupapi.h>
#include <stdio.h>
#include <hidsdi.h>
#include <conio.h>

/** local definitions **/
#define MYDEVICE_VID  0xcb00
#define MYDEVICE_PID  0x0005
/* default settings */

/** external functions **/

/** external data **/

/** internal functions **/
int FindSuitableHIDDevice (PCHAR *ppFoundDeviceName);

/** public data **/

/** private data **/

/** public functions **/

/** private functions **/


int main(int argc, char* argv[])
{
	PCHAR pMyDevice = NULL;
	HANDLE hHidDevice;

	if (FindSuitableHIDDevice (&pMyDevice) == 0)
		{
			printf ("Device Found\n\n");
			if (pMyDevice != NULL)
				{
					PHIDP_PREPARSED_DATA pHIDPreparsedData;
					BOOL allDone = FALSE;
					HIDP_CAPS hidCaps;

					/* This causes access denied on some Win2K machines 
					hHidDevice = CreateFile (pMyDevice,
							                             GENERIC_READ | GENERIC_WRITE,
		                                       FILE_SHARE_READ	| FILE_SHARE_WRITE,
		                                       NULL,        // no SECURITY_ATTRIBUTES structure
		                                       OPEN_EXISTING, // No special create flags
		                                       0, 
		                                       NULL);       // No template file
					*/
					hHidDevice = CreateFile (pMyDevice,
							                             0,
		                                       FILE_SHARE_READ	| FILE_SHARE_WRITE,
		                                       NULL,        // no SECURITY_ATTRIBUTES structure
		                                       OPEN_EXISTING, // No special create flags
		                                       0, 
		                                       NULL);       // No template file
					if (hHidDevice == INVALID_HANDLE_VALUE)
						{
							printf ("Could not open handle to the HID device, Error: %d\n", GetLastError());
							return(1);
						}

					if (!HidD_GetPreparsedData (hHidDevice, &pHIDPreparsedData))
						{
							printf ("Could not get Preparsed Data\n");
						}
					else
						{
							if (HidP_GetCaps (pHIDPreparsedData, &hidCaps) != HIDP_STATUS_SUCCESS)
								printf ("Could not GetCaps on preparsed data\n");
							else
								{
									printf ("HID Caps: \n");
									printf ("InputReportByteLength = %d\n", hidCaps.InputReportByteLength);
									printf ("OutputReportByteLength = %d\n", hidCaps.OutputReportByteLength);
									printf ("FeatureReportByteLength = %d\n", hidCaps.FeatureReportByteLength);
								}

						}


					//Ask the user what should be done

					while (!allDone)
						{
							char choice;
							int value;
							char repBuffer[2];
							//DWORD numBytesRead, numBytesWritten;

							printf ("R to read, W to Write, Q to quit: ");
							choice = getchar ();
							//Need to discard the CR
							getchar ();


							switch (choice)
								{
									case 'r':
									case 'R':
										//ToDo: Add code to read a report
										repBuffer[0] = 0;
										if (HidD_GetFeature (hHidDevice, repBuffer, sizeof (repBuffer)))
										//if (ReadFile (hHidDevice, repBuffer, 2, &numBytesRead, NULL))
											printf ("Value Read = %02x\n", (BYTE)repBuffer[1]);
										else
											printf ("Error Reading %d!\n", GetLastError());
										break;
									case 'w':
									case 'W':
										printf ("Enter Byte Value to Write: ");
										scanf ("%d", &value);
										//Need to discard the CR
										getchar ();;
										printf ("Will Write = %02x\n", (BYTE)value);
										//ToDo: Add code to write a report
										repBuffer[0] = 0;
										repBuffer[1] = (char)value;
										//if (WriteFile (hHidDevice, repBuffer, sizeof (repBuffer), &numBytesWritten, NULL))
										if (HidD_SetFeature (hHidDevice, repBuffer, sizeof (repBuffer)))
											printf ("Value Written = %02x\n", (BYTE)value);
										else
											printf ("Error Writing = %d!\n", GetLastError());
										break;
									case 'q':
									case 'Q':
										allDone = TRUE;
										break;
								}
						}

					CloseHandle (hHidDevice);

				}
			else
				printf ("Device Path was not found!\n");
		}
	else
		printf ("Device Not Found\n");

	return 0;
}



/*-------------------------------------------------------------
* FindSuitableHIDDevice
*
* Parameters: ppFoundDeviceName -> Pointer to a Pointer to hold the new device path
*
* Description: This routine tries to locate the HID device.
*
* Globals Affected/Used: None
*
* Returns: 0 on success, -1 on error
*
*-------------------------------------------------------------*/
int FindSuitableHIDDevice (PCHAR *ppFoundDeviceName)
	{
		GUID hidClassGUID;
		HDEVINFO hDeviceInfo = INVALID_HANDLE_VALUE;
		SP_DEVICE_INTERFACE_DATA deviceInterfaceData;
		UINT currentHidDevice, deviceInterfaceDetailLength;
		PSP_INTERFACE_DEVICE_DETAIL_DATA pDeviceInterfaceDetail = NULL;
		HIDD_ATTRIBUTES hidDeviceAttributes;
		HANDLE hHidDevice = INVALID_HANDLE_VALUE;
		int retVal = -1;
		
		*ppFoundDeviceName = NULL;
    HidD_GetHidGuid (&hidClassGUID);

    //
    // Open a handle to the plug and play dev node.
    //
    hDeviceInfo = SetupDiGetClassDevs ( &hidClassGUID,
                                               NULL, // Define no enumerator (global)
                                               NULL, // Define no
                                               (DIGCF_PRESENT | // Only Devices present
                                                DIGCF_DEVICEINTERFACE)); // Function class devices.
		if (hDeviceInfo == INVALID_HANDLE_VALUE)
			return(-1);

		deviceInterfaceData.cbSize = sizeof (SP_INTERFACE_DEVICE_DATA);
		currentHidDevice = 0;
		while (SetupDiEnumDeviceInterfaces (hDeviceInfo, 0, &hidClassGUID, currentHidDevice, 
			&deviceInterfaceData))
			{
				hHidDevice = INVALID_HANDLE_VALUE;

				//printf ("First Device\n");

				//First find out the size of data that's coming back...
				SetupDiGetDeviceInterfaceDetail (hDeviceInfo, &deviceInterfaceData, NULL, 0,
					&deviceInterfaceDetailLength, NULL);

				//printf ("deviceInterfaceDetailLength = %08d\n", deviceInterfaceDetailLength);

				pDeviceInterfaceDetail = malloc (deviceInterfaceDetailLength);
				if (pDeviceInterfaceDetail == NULL)
					goto next_device;

				pDeviceInterfaceDetail->cbSize = sizeof (SP_INTERFACE_DEVICE_DETAIL_DATA);
				if (!SetupDiGetDeviceInterfaceDetail (hDeviceInfo, &deviceInterfaceData, pDeviceInterfaceDetail, 
					deviceInterfaceDetailLength, &deviceInterfaceDetailLength, NULL))
					{
						printf ("Could not GetDeviceInterfaceDetail\n");
						goto next_device;
					}


				hHidDevice = CreateFile (pDeviceInterfaceDetail->DevicePath,
                                       0,
                                       FILE_SHARE_READ	| FILE_SHARE_WRITE,
                                       NULL,        // no SECURITY_ATTRIBUTES structure
                                       OPEN_EXISTING, // No special create flags
                                       0, 
                                       NULL);       // No template file

				if (hHidDevice != INVALID_HANDLE_VALUE)
					{
						if (HidD_GetAttributes (hHidDevice, &hidDeviceAttributes))
							{
								//printf ("VendorID = %04x, ProductID = %04x\n", hidDeviceAttributes.VendorID,
								//		hidDeviceAttributes.ProductID);
								if ((hidDeviceAttributes.VendorID != MYDEVICE_VID) ||
									(hidDeviceAttributes.ProductID != MYDEVICE_PID))
									goto next_device;
								else
									{
										*ppFoundDeviceName = malloc (lstrlen (pDeviceInterfaceDetail->DevicePath));
										lstrcpy (*ppFoundDeviceName, pDeviceInterfaceDetail->DevicePath);
										retVal = 0;
										break;
									}
							}
						else
							printf ("Could not GetAttributes\n");
						CloseHandle (hHidDevice);
				}

next_device:
				free (pDeviceInterfaceDetail);
				pDeviceInterfaceDetail = NULL;
				currentHidDevice++;
			}

		//We're done with device info stuff
		if (pDeviceInterfaceDetail)
			{
				free (pDeviceInterfaceDetail);
				pDeviceInterfaceDetail = NULL;
			}

		SetupDiDestroyDeviceInfoList (hDeviceInfo);

		return retVal;
	}


/*----------------------------------------------------------------------------------
* $Log: hosttest.c,v $
* Revision 1.5  2002/10/30 21:19:27  Devendra
* Removed Read and Write access from CreateFile to allow ioctl only access to the HID device.
*
* Revision 1.4  2002/10/30 02:36:13  Devendra
* - Minor change to hosttest to report error codes if device handle does not open.
*
* Revision 1.3  2002/10/22 17:50:59  Devendra
* - Changed Product ID to 0x0005
*
* Revision 1.2  2002/10/22 17:47:18  Devendra
* - Refine prompts, displays.
* - Change output location.
*
* Revision 1.1  2002/10/22 17:23:38  Devendra
* Rearranged file locations.
*
* Revision 1.2  2002/10/14 05:16:04  Devendra
* - Custom Device now using Feature reports for I/O
* - Added compiler define/project configuration to switch between mouse and custom device.
*
* Revision 1.1  2002/10/14 01:33:15  Devendra
* Got "custom" device functional with buttons and lights!
*
*
*---------------------------------------------------------------------------------*/


